##
##  UCS / The R ToolBox
##

## Filename: plots.R
## Modified: Sun Feb 29 15:13:30 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: High-level commands for precision and recall plots


ucs.library("base")
ucs.library("sfunc")


## PAR$col     = vector of colours for precision graphs (etc.)
## PAR$lty     = vector of line types for precision graphs
## PAR$lwd     = vector of line widths of precision graphs
## PAR$bw.col  = b/w mode: vector of colours for precision graphs
## PAR$bw.lty  = b/w mode: vector of line types for precision graphs
## PAR$bw.lwd  = b/w mode: vector of line widths of precision graphs
## PAR$n.first = smallest n for which n-best list is computed (default: 100)
## PAR$n.step  = step width for n-best lists (default: 1)
## PAR$tp.first = smallest number of TPs for which recall/precision is shown (default: 40)
## PAR$tp.step  = step width for recall values (number of TPs, default: 1)
## PAR$cex     = character expansion factor (passed to whatever output device a plot function uses)
##   global configuration parameters for precision and recall plots
##   (can be modified directly by assigning to the parameters; reload "plots.R" to recover defaults)
PAR <- list(col = rep(c("#000000", "#FF0000", "#0000FF", "#008800", "#FF6600"), 2),
            lty = c(rep("solid", 5), rep("dashed", 5)),
            lwd = rep(2, 10),
            bw.col = c(rep("black", 4), rep("#666666", 4), "black", "#666666"),
            bw.lty = c("solid","33","11","dotdash", "solid","33","11","dotdash", "longdash","longdash"),
            bw.lwd = c(2,3,4,3, 3,3,4,3, 3,3),
            n.first = 100,
            n.step = 1,
            tp.first = 40,
            tp.step = 1,
            cex = 1)


## check.pars()
##   used internally to ensure that PAR settings are sensible
check.pars <- function ( ) {
  PAR$col <<- rep(PAR$col, length=10)
  PAR$lty <<- rep(PAR$lty, length=10)
  PAR$lwd <<- rep(PAR$lwd, length=10)
  PAR$bw.col <<- rep(PAR$bw.col, length=10)
  PAR$bw.lty <<- rep(PAR$bw.lty, length=10)
  PAR$bw.lwd <<- rep(PAR$bw.lwd, length=10)
  # might run some sanity checks here
}



## precision.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
##   plot combined precision graphs for up to 10 association measures
##   options:
##     y.max      = max. precision value shown on y axis (determines aspect of view)
##     y.min,     = select plot region (for zooming in)
##     x.min, x.max
##     n.first    = overrides PAR$n.first (smallest n for which n-best list is computed)
##     n.step     = overrieds PAR$n.step (step width for n-best lists)
##     show.baseline = show baseline precision? (default: TRUE)
##     show.nbest = mark these n-best lists with vertical lines
##     show.npair = show as number of pair types, or nothing for show.ncand=NULL (default: size of data set)
##     title      = main title of plot (optional)
##     test       = test significance of differences, argument specifies sig. level (default: 0.05 for test=TRUE)
##     test.am1   = first AM in test  (H1 = AM1 is better than AM2, i.e. two-sided test)
##     test.am2   = second AM in test
##     test.step  = step size for tests (multiples of n.step; default: 10)
##     test.relevance = (absolute) difference precision that is considered relevant (percentage points)
##     conf       = show confidence region, argument specifies sig. level (default: 0.05 = 95% confidence for conf=TRUE)
##     conf.am    = AM for which confidence region is shown
##     cut        = consider first <cut> candidates from each ranking only (but show as 100%)
##     file       = save plot to EPS file (does not display on screen), argument specifies filename
##     aspect     = when saving to EPS file, enforce this aspect between y- and x-percentages (defaults to square plot)
##     cex        = character expansion factor for labels and legend box (default: 1.0)
##     bw         = display in b/w mode? (default: FALSE)
##     usercode   = callback function which can be used to add something to the plot before the legend box is drawn;
##                  invoked with parameters region=c(x.min, x.max, y.min, y.max) and
##                  pr=<list of precision.recall() data for selected AMs, indexed by AM key>
##   [further named parameters are interpreted as local graphics settings and are passed to the par() function]
##   [names of AMs for significance test and confidence region may be abbreviated to unique prefixes]
precision.plot <- function (ds, keys,
                            ...,        # extra arguments are interpreted as graphics parameters
                            y.max=100, x.max=100, y.min=0, x.min=0,
                            n.first=PAR$n.first, n.step=PAR$n.step,
                            show.baseline=TRUE,
                            show.nbest=NULL,
                            show.npair=NULL,
                            test=FALSE, test.am1=NULL, test.am2=NULL, test.step=10, test.relevant=0,
                            conf=FALSE, conf.am=NULL,
                            cut=NULL,
                            bw=FALSE,
                            usercode=NULL,
                            file=NULL, aspect=1,
                            cex=PAR$cex,
                            title=NULL
                            ) {
  delta.x <- x.max - x.min
  delta.y <- y.max - y.min

  ## if parameter <file> is specified, write plot to EPS file 
  if (! missing(file)) {
    ## choose height such that plot region has format delta.x by delta.y*aspect
    ## (assuming 20% margins around the x-axis, and a y-axis margin of 30% of the total x-axis dimension)
    ## 0.8 * width / delta.x = (height - 0.3*width) / (delta.y * aspect)
    ## height = (0.8 * (delta.y * aspect) / delta.x + 0.3) * width
    dev.width <- 6
    if (missing(aspect)) {
      dev.height <- dev.width
    } else {
      dev.height <- (0.8 * (delta.y * aspect) / delta.x + 0.3) * dev.width
    }
    postscript(file=file, width=dev.width, height=dev.height,
               paper="special", horizontal=FALSE, onefile=FALSE)
  }

  ## set optional "local" graphics parameters
  par.save  <- par(list(...))
  on.exit(par(par.save))                # arrange for original parameter values to be restored

  par(cex=cex)
  
  ## might automatically determine y.max here (i.e. pre-calculate pr.data)

  ## read default line styles (colour or b/w)
  check.pars()
  if (bw) {
    col <- PAR$bw.col
    lty <- PAR$bw.lty
    lwd <- PAR$bw.lwd
  } else {
    col <- PAR$col
    lty <- PAR$lty
    lwd <- PAR$lwd
  }

  ## set default significance levels for difference test and confidence region
  if (test==TRUE) test <- 0.05
  if (conf==TRUE) conf <- 0.05

  ## if user has asked for significance tests or confidence region, check that
  ## AMs are specified and resolve them in <keys> (using partial matching if necessary)
  if (conf) {
    if (missing(conf.am)) {
      stop("Missing argument conf.am for confidence region.")
    }
    idx <- pmatch(conf.am, keys)
    if (is.na(idx)) {
      stop(paste("Can't resolve conf.am=<",conf.am,"> in list of AM keys.", sep=""))
    }
    else {
      conf.am <- keys[idx]
    }
  }
  if (test) {
    if (missing(test.am1)) {
      stop("Missing argument test.am1 for significance tests.")
    }
    if (missing(test.am2)) {
      stop("Missing argument test.am2 for significance tests.")
    }
    idx.am1 <- pmatch(test.am1, keys)
    if (is.na(idx.am1)) {
      stop(paste("Can't resolve test.am1=<",test.am1,"> in list of AM keys.", sep=""))
    }
    else {
      test.am1 <- keys[idx.am1]
    }
    idx.am2 <- pmatch(test.am2, keys)
    if (is.na(idx.am2)) {
      stop(paste("Can't resolve test.am2=<",test.am2,"> in list of AM keys.", sep=""))
    }
    else {
      test.am2 <- keys[idx.am2]
    }
  }

  n.keys <- length(keys)                # number of AMs in this plot

  ## precompute and save precision and recall data for further tasks (incl. sig. tests and conf. region)
  pr.data <- list()
  for (i in 1:n.keys) {
    ## compute precision and recall values for current AM (with optional cut)
    if (missing(cut)) {
      pr <- precision.recall(ds, keys[i], first=n.first, step=n.step)
    } else {
      pr <- precision.recall(ds, keys[i], first=n.first, step=n.step, cut=cut)
    }
    pr.data[[ keys[i] ]]  <- pr
  }

  ## set up plot region and axes without plotting any curves
  xlab.text <- "proportion of ranked data set (%)"
  n.pair <- max(pr.data[[1]]$n)         # not the most efficient way to find out ...
  if (! is.null(show.npair)) {
    n.pair <- show.npair                # override for evaluation of random samples
  }
  if (missing(show.npair) || (! is.null(show.npair))) {
    ## show number of pair types unless show.ncand=NULL was specified
    xlab.text <- paste(xlab.text, "[of", n.pair, "pair types]")
  }
  plot(0,0, type="n", xlim=c(x.min, x.max), ylim=c(y.min, y.max),
       xlab=xlab.text, ylab="precision (%)",
       xaxs="i", yaxs="i",
       main=title)

  ## show confidence region as grey area in background
  if (conf) {
    pr <- pr.data[[conf.am]]            # retrieve precision/recall data for the requested AM
    upper <- 100 * binom.conf.interval(pr$tp, pr$n, limit="upper", conf.level=conf)
    lower <- 100 * binom.conf.interval(pr$tp, pr$n, limit="lower", conf.level=conf)
    if (bw) {
      polygon(c(pr$perc, rev(pr$perc)), c(upper, rev(lower)), # encircle conf. region outline in clockwise direction
              density=10, col="black", border="black") # shade region diagonally in b/w mode
    } else {
      polygon(c(pr$perc, rev(pr$perc)), c(upper, rev(lower)), 
              col="#AAAAAA", border="black") # draw region in light grey with black outline in colour mode
    }
  }

  if (show.baseline) {
    ## baseline precision (percentage)
    baseline <- 100 * sum(ds$b.TP) / length(ds$b.TP)
    abline(h=baseline, lty="dotted", col="black", lwd=2)
    
    ## show baseline prec. value on the right (-5%) below the baseline (-3%)
    txt.x <- x.max - 0.05 * delta.x
    txt.y <- baseline - 0.02 * delta.y
    text(txt.x, txt.y, adj=c(1,1),
         labels=paste("baseline = ", round(baseline, dig=2), "%", sep=""))
  }
  
  ## plot precision graphs for all AMs
  for (i in 1:n.keys) {
    pr <- pr.data[[ keys[i] ]]          # retrieve precision/recall data for current AM
    ## plot precision values with corresponding lty and col
    points(pr$perc, pr$precision, type="l", lwd=lwd[i], lty=lty[i], col=col[i])
  }

  ## if show.nbest was specified, indicate the requested n-best lists by vertical lines
  if (! is.null(show.nbest)) {
    for (n in show.nbest) {
      pr <- pr.data[[1]]                # get corresponding x-axis location from any one of the precision/recall tables
      idx <- sum(pr$n <= n)             # this computes the largest row in pr for which the <n> column value is <= n
      abline(v=pr$perc[idx], lwd=1, col="black")
    }
  }
  
  ## perform tests for significance of differences between two measures
  yellow.triangle <- if (!bw) "yellow" else "#AAAAAA"  # define fill colour for yellow & red triangles
  red.triangle <- if (!bw) "red" else "black"
  if (test) {
    steps <- length(pr.data[[1]]$n)     # perform tests for subset of n-best lists shown in graphs
    rank.am1 <- ds[[ am.key2rvar(test.am1) ]] # need rankings of the TPs for McNemar's test
    rank.am2 <- ds[[ am.key2rvar(test.am2) ]]
    tp <- as.logical(ds$b.TP)           # make sure it's a logical vector (for subsetting)
    rank.am1 <- rank.am1[tp]
    rank.am2 <- rank.am2[tp]
    for (i in seq(1, steps, by=test.step)) {
      n <- pr.data[[test.am1]]$n[i]     # use same n-best lists as for the curves
      x <- pr.data[[test.am1]]$perc[i]  # x-coordinate and y-coordinates for both AMs (precision) for plotting
      y1 <- pr.data[[test.am1]]$precision[i]
      y2 <- pr.data[[test.am2]]$precision[i]
      ## run exact McNemar's test (H0 = differences are random, H1 = AM1 has more TPs in n-best list)
      k.1  <- sum(rank.am1 <= n & rank.am2 > n) # how often AM1 made the better choice
      k.2  <- sum(rank.am1 > n & rank.am2 <= n) # how often AM2 made the better choice
      k <- k.1 + k.2                    # total number of differences
      if (k > 0) {
        p <- pbinom(k.1 - 1, prob=0.5, size=k, lower=FALSE)
        if (p < test) {                 # yellow triangle = significant evidence for some difference
          points(x, y1, col="black", bg=yellow.triangle, pch=25) # down triangle for AM1
          points(x, y2, col="black", bg=yellow.triangle, pch=24) # up triangle for AM2
        }
        if (test.relevant) {            # red triangle = significant evidence for relevant difference 
          k.req <- test.relevant * n / 100  # AM1 must make better choice for k.req TPs for relevant difference
          p.req <- (k.req / k + 1) / 2  # this is achieved on average if success probability in McNemar's test is >= p.req
          if (p.req <= 1) {            
            p <- pbinom(k.1 - 1, prob=p.req, size=k, lower=FALSE)
            if (p < test) {
              points(x, y1, col="black", bg=red.triangle, pch=25)
              points(x, y2, col="black", bg=red.triangle, pch=24)
            }
          }
        }
      }
    }
  }

  ## execute user callback function, which can add something to the plot before it is written to an EPS file
  if (! is.null(usercode)) {
    usercode(region=c(x.min, x.max, y.min, y.max), pr=pr.data)
  }
  
  ## simple legend box (move in from upper right corner by 5%)
  leg.x <- x.max - 0.05 * delta.x
  leg.y <- y.max - 0.05 * delta.y
  leg.pch  <- rep(-1, n.keys)           # if sig.test results are shown, display corresp. triangles in legend
  leg.bg <- rep("black", n.keys)        # background colours for triangle symbols
  if (test) {
    leg.pch[c(idx.am1, idx.am2)] <- c(25,24)
    leg.bg[c(idx.am1, idx.am2)] <- rep(yellow.triangle, 2)
  }
  legend(leg.x, leg.y, xjust=1, yjust=1,# top right corner
         legend=keys,                   # show AM keys in legend
         pch=leg.pch, pt.bg=leg.bg,
         lwd=3, lty=lty[1:n.keys], col=col[1:n.keys],
         bg="white")

  ## if writing to EPS file, close file and associated device now
  if (! missing(file)) {
    dev.off()                           # close EPS file device
  }
}



## recall.precision.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
##   plot recall-against-precision graphs for up to 10 association measures
##   options:
##     y.max      = max. precision value shown on y axis (determines aspect of view)
##     y.min,     = select plot region (for zooming in)
##     x.min, x.max
##     tp.first   = overrides PAR$tp.first (smallest number of TPs for which recall/precision is shown)
##     tp.step    = overrieds PAR$tp.step (step width for recall values, number of TPs)
##     show.baseline = show baseline precision? (default: TRUE)
##     show.recall = mark these recall percentages with vertical lines
##     show.nbest = indicate these n-best lists with diagonal lines
##     show.ntp   = show total number of TPs (on axis label)
##     title      = main title of plot (optional)
##     file       = save plot to EPS file (does not display on screen), argument specifies filename
##     aspect     = when saving to EPS file, enforce this aspect between y- and x-percentages (defaults to square plot)
##     cex        = character expansion factor for labels and legend box (default: 1.0)
##     bw         = display in b/w mode? (default: FALSE)
##   [further named parameters are interpreted as local graphics settings and are passed to the par() function]
recall.precision.plot <- function (ds, keys,
                                   ...,        # extra arguments are interpreted as graphics parameters
                                   y.max=100, x.max=100, y.min=0, x.min=0,
                                   tp.first=PAR$tp.first, tp.step=PAR$tp.step,
                                   show.baseline=TRUE,
                                   show.recall=NULL,
                                   show.nbest=NULL,
                                   show.ntp=FALSE,
                                   bw=FALSE,
                                   file=NULL, aspect=1,
                                   cex=PAR$cex,
                                   title=NULL
                                   ) {
  delta.x <- x.max - x.min
  delta.y <- y.max - y.min

  ## if parameter <file> is specified, write plot to EPS file 
  if (! missing(file)) {
    ## choose height such that plot region has format delta.x by delta.y*aspect
    ## (assuming 20% margins around the x-axis, and a y-axis margin of 30% of the total x-axis dimension)
    ## 0.8 * width / delta.x = (height - 0.3*width) / (delta.y * aspect)
    ## height = (0.8 * (delta.y * aspect) / delta.x + 0.3) * width
    dev.width <- 6
    if (missing(aspect)) {
      dev.height <- dev.width
    } else {
      dev.height <- (0.8 * (delta.y * aspect) / delta.x + 0.3) * dev.width
    }
    postscript(file=file, width=dev.width, height=dev.height,
               paper="special", horizontal=FALSE, onefile=FALSE)
  }

  ## set optional "local" graphics parameters
  par.save  <- par(list(...))
  on.exit(par(par.save))                # arrange for original parameter values to be restored

  par(cex=cex)

  ## might automatically determine y.max here (i.e. pre-calculate pr.data)

  ## read default line styles (colour or b/w)
  check.pars()
  if (bw) {
    col <- PAR$bw.col
    lty <- PAR$bw.lty
    lwd <- PAR$bw.lwd
  } else {
    col <- PAR$col
    lty <- PAR$lty
    lwd <- PAR$lwd
  }

  n.keys <- length(keys)                # number of AMs in this plot

  ## precompute and save precision and recall data for further tasks (incl. sig. tests and conf. region)
  pr.data <- list()
  for (i in 1:n.keys) {
    pr.data[[ keys[i] ]]  <- precision.recall(ds, keys[i])
  }

  n.tp <- max(pr.data[[1]]$tp)          # total number of TPs

  ## set up plot region and axes without plotting any curves
  xlab.text <- "recall (%)"
  if (show.ntp) xlab.text <- paste(xlab.text, "[of", n.tp, "TPs]")
  plot(0,0, type="n", xlim=c(x.min, x.max), ylim=c(y.min, y.max),
       xlab=xlab.text, ylab="precision (%)",
       xaxs="i", yaxs="i",
       main=title)

  if (show.baseline) {
    ## baseline precision (percentage)
    baseline <- 100 * sum(ds$b.TP) / length(ds$b.TP)
    abline(h=baseline, lty="dotted", col="black", lwd=2)
    
    ## show baseline prec. value on the right (-5%) below the baseline (-3%)
    txt.x <- x.max - 0.05 * delta.x
    txt.y <- baseline - 0.02 * delta.y
    text(txt.x, txt.y, adj=c(1,1),
         labels=paste("baseline = ", round(baseline, dig=2), "%", sep=""))
  }
  
  ## plot precision against recall for all AMs
  for (i in 1:n.keys) {
    pr <- pr.data[[ keys[i] ]]          # retrieve precision/recall data for current AM
    ## for a given recall value, only show the first n-best list achieving it -- i.e. we
    ## need to identify (starting points of) RUNs in the vector of TP counts as efficiently as possible
    tp <- pr$tp                         # vector of TP counts
    n <- length(tp)
    tp.shift <- c(0, pr$tp[-n])         # shifted vector, hence the starting point of a RUN satisfies tp != tp.shift
    idx <- (tp != tp.shift) & (tp >= tp.first) # RUN start points with number of TPs >= tp.first
    if (! missing(tp.step)) {            # if step was specified, force number of TPs to be a multiple of step, starting from recall=100%
      idx <- idx & ((n.tp - tp) %% tp.step == 0)
    }
    points(pr$recall[idx], pr$precision[idx], type="l", lwd=lwd[i], lty=lty[i], col=col[i])
  }

  ## if show.recall was specified, indicate the requested recall percentages by vertical lines
  if (! is.null(show.recall)) {
    for (r in show.recall) {
      abline(v=r, lwd=1, col="black")
    }
  }
  
  ## if show.nbest was specified, indicate the requested n-best lists with diagonal lines
  if (! is.null(show.nbest)) {
    for (n in show.nbest) {
      slope <- n.tp / n
      abline(0, slope, lwd=1, col="black")
    }
  }
  
  ## simple legend box (move in from upper right corner by 5%)
  leg.x <- x.max - 0.05 * delta.x
  leg.y <- y.max - 0.05 * delta.y
  legend(leg.x, leg.y, xjust=1, yjust=1,# top right corner
         legend=keys,                   # show AM keys in legend
         lwd=3, lty=lty[1:n.keys], col=col[1:n.keys],
         bg="white")

  ## if writing to EPS file, close file and associated device now
  if (! missing(file)) {
    dev.off()                           # close EPS file device
  }
}



## recall.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
##   plot combined recall graphs for up to 10 association measures
##   options:
##     n.step     = overrieds PAR$n.step (step width for n-best lists)
##     show.nbest = mark these n-best lists with vertical lines
##     show.npair = show as number of pair types, or nothing for show.ncand=NULL (default: size of data set)
##     title      = main title of plot (optional)
##     cut        = consider first <cut> candidates from each ranking only (but show as 100%)
##     file       = save plot to EPS file (does not display on screen), argument specifies filename
##     cex        = character expansion factor for labels and legend box (default: 1.0)
##     bw         = display in b/w mode? (default: FALSE)
##   [further named parameters are interpreted as local graphics settings and are passed to the par() function]
recall.plot <- function (ds, keys,
                         ...,        # extra arguments are interpreted as graphics parameters
                         n.first=PAR$n.first, n.step=PAR$n.step,
                         show.nbest=NULL,
                         show.npair=NULL,
                         cut=NULL,
                         bw=FALSE,
                         file=NULL, aspect=1,
                         cex=PAR$cex,
                         title=NULL
                         ) {

  ## if parameter <file> is specified, write plot to EPS file 
  if (! missing(file)) {
    dev.width <- 6
    dev.height <- dev.width
    postscript(file=file, width=dev.width, height=dev.height,
               paper="special", horizontal=FALSE, onefile=FALSE)
  }

  ## set optional "local" graphics parameters
  par.save  <- par(list(...))
  on.exit(par(par.save))                # arrange for original parameter values to be restored

  par(cex=cex)

  ## read default line styles (colour or b/w)
  check.pars()
  if (bw) {
    col <- PAR$bw.col
    lty <- PAR$bw.lty
    lwd <- PAR$bw.lwd
  } else {
    col <- PAR$col
    lty <- PAR$lty
    lwd <- PAR$lwd
  }

  n.keys <- length(keys)                # number of AMs in this plot

  ## set up plot region and axes without plotting any curves
  xlab.text <- "proportion of ranked data set (%)"
  n.pair <- nrow(ds)
  if (! is.null(show.npair)) {
    n.pair <- show.npair                # override for evaluation of random samples
  }
  if (missing(show.npair) || (! is.null(show.npair))) {
    ## show number of pair types unless show.ncand=NULL was specified
    xlab.text <- paste(xlab.text, "[of", n.pair, "pair types]")
  }
  plot(0,0, type="n", xlim=c(0, 100), ylim=c(0, 100),
       xlab=xlab.text, ylab="recall (%)",
       xaxs="i", yaxs="i",
       main=title)
  
  ## plot recall graphs for all AMs
  for (i in 1:n.keys) {
    pr <- precision.recall(ds, keys[i], step=n.step)
    points(pr$perc, pr$recall, type="l", lwd=lwd[i], lty=lty[i], col=col[i])
  }

  ## if show.nbest was specified, indicate the requested n-best lists by vertical lines
  if (! is.null(show.nbest)) {
    for (n in show.nbest) {
      idx <- sum(pr$n <= n)             # pr still holds precision/recall data from last AM in list
      abline(v=pr$perc[idx], lwd=1, col="black")
    }
  }
  
  ## simple legend box (move in from lower right corner by 5%)
  legend(95, 05, xjust=1, yjust=0,
         legend=keys,                   # show AM keys in legend
         lwd=3, lty=lty[1:n.keys], col=col[1:n.keys],
         bg="white")

  ## if writing to EPS file, close file and associated device now
  if (! missing(file)) {
    dev.off()                           # close EPS file device
  }
}



## local.precision.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
##   plot combined local precision graphs for up to 10 association measures
##   options:
##     y.max      = max. precision value shown on y axis (determines aspect of view)
##     y.min,     = select plot region (for zooming in)
##     x.min, x.max
##     n.first    = overrides PAR$n.first (smallest n for which n-best list is computed)
##     n.step     = overrides PAR$n.step (step width for n-best lists)
##     window     = window size for local precision (default: 100; number of pairs in window is 2*<window>+1)
##     show.baseline = show baseline precision? (default: TRUE)
##     show.nbest = mark local precision around n-th pair type with vertical line
##     show.npair = show as number of pair types, or nothing for show.ncand=NULL (default: size of data set)
##     title      = main title of plot (optional)
##     cut        = consider first <cut> candidates from each ranking only (but show as 100%)
##     file       = save plot to EPS file (does not display on screen), argument specifies filename
##     aspect     = when saving to EPS file, enforce this aspect between y- and x-percentages (defaults to square plot)
##     cex        = character expansion factor for labels and legend box (default: 1.0)
##     bw         = display in b/w mode? (default: FALSE)
##     usercode   = callback function which can be used to add something to the plot before the legend box is drawn;
##                  invoked with parameters region=c(x.min, x.max, y.min, y.max) and
##                  pr=<list of precision.recall() data for selected AMs, indexed by AM key>
##   [further named parameters are interpreted as local graphics settings and are passed to the par() function]
##   [names of AMs for significance test and confidence region may be abbreviated to unique prefixes]
local.precision.plot <- function (ds, keys,
                                  ...,        # extra arguments are interpreted as graphics parameters
                                  y.max=100, x.max=100, y.min=0, x.min=0,
                                  n.first=PAR$n.first, n.step=PAR$n.step,
                                  window=100,
                                  show.baseline=TRUE,
                                  show.nbest=NULL,
                                  show.npair=NULL,
                                  cut=NULL,
                                  bw=FALSE,
                                  usercode=NULL,
                                  file=NULL, aspect=1,
                                  cex=PAR$cex,
                                  title=NULL
                                  ) {
  delta.x <- x.max - x.min
  delta.y <- y.max - y.min

  ## if parameter <file> is specified, write plot to EPS file 
  if (! missing(file)) {
    ## choose height such that plot region has format delta.x by delta.y*aspect
    ## (assuming 20% margins around the x-axis, and a y-axis margin of 30% of the total x-axis dimension)
    ## 0.8 * width / delta.x = (height - 0.3*width) / (delta.y * aspect)
    ## height = (0.8 * (delta.y * aspect) / delta.x + 0.3) * width
    dev.width <- 6
    if (missing(aspect)) {
      dev.height <- dev.width
    } else {
      dev.height <- (0.8 * (delta.y * aspect) / delta.x + 0.3) * dev.width
    }
    postscript(file=file, width=dev.width, height=dev.height,
               paper="special", horizontal=FALSE, onefile=FALSE)
  }

  ## set optional "local" graphics parameters
  par.save  <- par(list(...))
  on.exit(par(par.save))                # arrange for original parameter values to be restored

  par(cex=cex)
  
  ## might automatically determine y.max here (i.e. pre-calculate pr.data)

  ## read default line styles (colour or b/w)
  check.pars()
  if (bw) {
    col <- PAR$bw.col
    lty <- PAR$bw.lty
    lwd <- PAR$bw.lwd
  } else {
    col <- PAR$col
    lty <- PAR$lty
    lwd <- PAR$lwd
  }

  n.keys <- length(keys)                # number of AMs in this plot

  ## precompute and save precision and recall data for further tasks (incl. sig. tests and conf. region)
  pr.data <- list()
  for (i in 1:n.keys) {
    ## compute precision and recall values for current AM (with optional cut)
    if (missing(cut)) {
      pr <- precision.recall(ds, keys[i], first=n.first, step=n.step, window=window)
    } else {
      pr <- precision.recall(ds, keys[i], first=n.first, step=n.step, window=window, cut=cut)
    }
    pr.data[[ keys[i] ]]  <- pr
  }

  ## set up plot region and axes without plotting any curves
  xlab.text <- "proportion of ranked data set (%)"
  max.n <- max(pr.data[[1]]$n)         # not the most efficient way to find out ...
  n.pair <- max.n
  if (! is.null(show.npair)) {
    n.pair <- show.npair                # override for evaluation of random samples
  }
  if (missing(show.npair) || (! is.null(show.npair))) {
    ## show number of pair types unless show.ncand=NULL was specified
    xlab.text <- paste(xlab.text, "[of", n.pair, "pair types]")
  }
  ylab.text <- paste("local precision (%) [W=", window, "]", sep="")
  plot(0,0, type="n", xlim=c(x.min, x.max), ylim=c(y.min, y.max),
       xlab=xlab.text, ylab=ylab.text,
       xaxs="i", yaxs="i",
       main=title)

  if (show.baseline) {
    ## baseline precision (percentage)
    baseline <- 100 * sum(ds$b.TP) / length(ds$b.TP)
    abline(h=baseline, lty="dotted", col="black", lwd=2)
    
    ## show baseline prec. value on the right (-5%) below the baseline (-3%)
    txt.x <- x.max - 0.05 * delta.x
    txt.y <- baseline - 0.02 * delta.y
    text(txt.x, txt.y, adj=c(1,1),
         labels=paste("baseline = ", round(baseline, dig=2), "%", sep=""))
  }
  
  ## plot local precision graphs for all AMs
  for (i in 1:n.keys) {
    pr <- pr.data[[ keys[i] ]]          # retrieve precision/recall data for current AM
    idx <- (pr$n > window) & (pr$n <= (max.n - window)) # remove marginal points where local precision is not defined 
    ## plot local precision values with corresponding lty and col
    points(pr$perc[idx], pr$lp[idx], type="l", lwd=lwd[i], lty=lty[i], col=col[i])
  }

  ## if show.nbest was specified, indicate the requested n-best lists by vertical lines
  if (! is.null(show.nbest)) {
    for (n in show.nbest) {
      pr <- pr.data[[1]]                # get corresponding x-axis location from any one of the precision/recall tables
      idx <- sum(pr$n <= n)             # this computes the largest row in pr for which the <n> column value is <= n
      abline(v=pr$perc[idx], lwd=1, col="black")
    }
  }
  
  ## execute user callback function, which can add something to the plot before it is written to an EPS file
  if (! is.null(usercode)) {
    usercode(region=c(x.min, x.max, y.min, y.max), pr=pr.data)
  }
  
  ## simple legend box (move in from upper right corner by 5%)
  leg.x <- x.max - 0.05 * delta.x
  leg.y <- y.max - 0.05 * delta.y
  legend(leg.x, leg.y, xjust=1, yjust=1,# top right corner
         legend=keys,                   # show AM keys in legend
         lwd=3, lty=lty[1:n.keys], col=col[1:n.keys],
         bg="white")

  ## if writing to EPS file, close file and associated device now
  if (! missing(file)) {
    dev.off()                           # close EPS file device
  }
}



